module.exports = function(RED) {
    "use strict";

    var dnp3 = require('dnp3lib');

    function DNP3OutstationNode(n) {
        RED.nodes.createNode(this, n);

        var node = this;

        let addrOutstation = Number(n.addrOutstation);
        let addrMaster = Number(n.addrMaster);
        let localAddr = n.localAddr || '0.0.0.0';
        let localPort = Number(n.localPort);
        let remotePort = Number(n.remotePort);
        let channel;

        switch (n.channel) {
            case 'tcp':
                if (n.tcp == 'server')
                    this.outstation = new dnp3.Outstation(addrOutstation, addrMaster, dnp3.channelTcp,
                                                          dnp3.tcpServer, localAddr, localPort);
                else
                    this.outstation = new dnp3.Outstation(addrOutstation, addrMaster, dnp3.channelTcp,
                                                          dnp3.tcpClient, n.remoteAddr, remotePort, localAddr);
                break;
            case 'udp':
                this.outstation = new dnp3.Outstation(addrOutstation, addrMaster, dnp3.channelUdp,
                                                      n.remoteAddr, remotePort, n.localAddr, localPort);
                break;
            case 'serial':
                this.outstation = new dnp3.Outstation(addrOutstation, addrMaster, dnp3.channelSerial,
                                                      n.device, Number(n.speed), Number(n.databits), dnp3[n.parity], dnp3[n.stopbits]);
                break;
        }

        this.outstation.channel.statusCallback = function (status) {
            node.emit('status', status);
        };

        this.outstation.channel.logCallback = function (level, message) {
            message = 'dnp3: ' + message;
            switch (level) {
                case dnp3.ERROR:
                    RED.log.error(message);
                    break;
                case dnp3.WARNING:
                    RED.log.warn(message);
                    break;
                case dnp3.INFO:
                    RED.log.info(message);
                    break;
                case dnp3.DEBUG:
                    RED.log.debug(message);
                    break;
                case dnp3.TRACE:
                    RED.log.trace(message);
                    break;
            }
        };

        this.outstation.keepalive = n.keepalive === '' ? null : Number(n.keepalive);
        this.outstation.unsolicitedAllow = !n.unsolDisable;
        this.outstation.solicitedTimeout = Number(n.solTimeout);
        this.outstation.unsolicitedTimeout = Number(n.unsolTimeout);
        this.outstation.unsolicitedRetries = n.unsolRetries === '' ? null : Number(n.unsolRetries);
        this.outstation.timeSync = Number(n.timesync);
        this.outstation.warmRestart = !!n.restartWarm;
        this.outstation.coldRestart = !!n.restartCold;

        this.on('close', () => {
            node.outstation.stop();
        });
    }

    function DNP3OutstationOutNode(n) {
        RED.nodes.createNode(this, n);

        var node = this;

        this.outstation = n.outstation;
        this.dataPoint = n.dataPoint;
        this.index = Number(n.index);
        this.mode = dnp3[n.mode];
        this.staticVariation = dnp3[n.staticVariation];
        this.eventVariation = dnp3[n.eventVariation];
        this.deadband = Number(n.deadband);
        this.clazz = dnp3[n.clazz];

        this.os = RED.nodes.getNode(this.outstation).outstation;

        switch (this.dataPoint) {
            case 'binary_in':
                this.os.addBinaryInput(this.staticVariation, this.eventVariation, this.clazz, this.index, this.mode);
                break;
            case 'dbbinary_in':
                this.os.addDoubleBinaryInput(this.staticVariation,this.eventVariation, this.clazz, this.index, this.mode);
                break;
            case 'counter_in':
                this.os.addCounter(this.staticVariation, this.eventVariation, this.clazz, this.deadband, this.index, this.mode);
                break;
            case 'analog_in':
                this.os.addAnalogInput(this.staticVariation, this.eventVariation, this.clazz, this.deadband, this.index, this.mode);
                break;
            case 'ostring_in':
                this.os.addOctetString(dnp3.group110var0, dnp3.group111var0, this.clazz, this.index, this.mode);
                break;
            case 'binary_out':
                if (this.os.binaryOutputs[this.index]) {
                    this.os.binaryOutputs[this.index].staticVariation = this.staticVariation;
                    this.os.binaryOutputs[this.index].eventVariation = this.eventVariation;
                    this.os.binaryOutputs[this.index].clazz = this.clazz;
                    this.os.binaryOutputs[this.index].mode = this.mode;
                } else {
                    this.os.addBinaryOutput(this.staticVariation, this.eventVariation, this.clazz, this.index, this.mode);
                }
                break;
            case 'analog_out':
                if (this.os.analogOutputs[this.index]) {
                    this.os.analogOutputs[this.index].staticVariation = this.staticVariation;
                    this.os.analogOutputs[this.index].eventVariation = this.eventVariation;
                    this.os.analogOutputs[this.index].clazz = this.clazz;
                    this.os.analogOutputs[this.index].deadband = this.deadband;
                    this.os.analogOutputs[this.index].mode = this.mode;
                } else {
                    this.os.addAnalogOutput(this.staticVariation, this.eventVariation, this.clazz, this.deadband, this.index, this.mode);
                }
                break;
        }

        node.on('input', (msg) => {
            switch (node.dataPoint) {
                case 'binary_in':
                    node.os.setBinaryInput(node.index, msg.payload, msg.quality, msg.timestamp);
                    break;
                case 'dbbinary_in':
                    node.os.setDoubleBinaryInput(node.index, msg.payload, msg.quality, msg.timestamp);
                    break;
                case 'counter_in':
                    node.os.setCounter(node.index, msg.payload, msg.quality, msg.timestamp);
                    break;
                case 'analog_in':
                    node.os.setAnalogInput(node.index, msg.payload, msg.quality, msg.timestamp);
                    break;
                case 'ostring_in':
                    node.os.setOctetString(node.index, msg.payload);
                    break;
                case 'binary_out':
                    node.os.setBinaryOutput(node.index, msg.payload, msg.quality, msg.timestamp);
                    break;
                case 'analog_out':
                    node.os.setAnalogOutput(node.index, msg.payload, msg.quality, msg.timestamp);
            }
        });

        RED.nodes.getNode(this.outstation).on('status', (status) => {
            switch (status) {
                case dnp3.channelClosed:
                    node.status({fill: 'red', shape: 'dot', text: 'disconnected'});
                    break;
                case dnp3.channelOpening:
                    node.status({fill: 'blue', shape: 'dot', text: 'connecting'});
                    break;
                case dnp3.channelOpened:
                    node.status({fill: 'green', shape: 'dot', text: 'connected'});
                    break;
                default:
                    node.status({fill: 'gray', shape: 'dot', text: 'unknow state'});
            }
        });

        node.on('close', () => {
            node.status({});
        });
    }

    function DNP3OutstationInNode(n) {
        RED.nodes.createNode(this, n);

        var node = this;

        this.outstation = n.outstation;
        this.dataPoint = n.dataPoint;
        this.analogVariation = dnp3[n.analogVariation];
        this.index = Number(n.index);

        this.os = RED.nodes.getNode(this.outstation).outstation;

        switch (this.dataPoint) {
            case 'binary_out':
                if (this.os.binaryOutputs[this.index]) {
                    this.os.binaryOutputs[this.index].callback = function(val) { node.emit('input', val); }
                } else {
                    this.os.addBinaryOutput(dnp3.group10var2, dnp3.group11var2, dnp3.class0, this.index, dnp3.modeStaticOnly, (val) => { node.emit('input', val); });
                }
                break;
            case 'analog_out':
                if (this.os.analogOutputs[this.index]) {
                    this.os.analogOutputs[this.index].callback = function(val) { node.emit('input', val); }
                } else {
                    this.os.addAnalogOutput(this.analogVariation, dnp3.group42var1, dnp3.class0, 0, this.index, dnp3.modeStaticOnly, (val) => { node.emit('input', val); });
                }
                break;
        }

        node.on('input', (val) => {
            let msg = {payload: val};
            this.send(msg);
        });

        RED.nodes.getNode(this.outstation).on('status', (status) => {
            switch (status) {
                case dnp3.channelClosed:
                    node.status({fill: 'red', shape: 'dot', text: 'disconnected'});
                    break;
                case dnp3.channelOpening:
                    node.status({fill: 'blue', shape: 'dot', text: 'connecting'});
                    break;
                case dnp3.channelOpened:
                    node.status({fill: 'green', shape: 'dot', text: 'connected'});
                    break;
                default:
                    node.status({fill: 'gray', shape: 'dot', text: 'unknow state'});
            }
        });

        node.on('close', () => {
            node.status({});
        });
    }

    RED.events.on('flows:started', () => {
        RED.nodes.eachNode((node) => {
            if (node.type == 'dnp3 outstation') {
                RED.nodes.getNode(node.id).outstation.start();
            }
        });
    });

    RED.nodes.registerType('dnp3 outstation', DNP3OutstationNode);
    RED.nodes.registerType('dnp3 outstation out', DNP3OutstationOutNode);
    RED.nodes.registerType('dnp3 outstation in', DNP3OutstationInNode);
}
